"""
    Copyright 2006,2007 Riccardo Taormina (den3b@hotmail.com)
    
    This file is part of SmartChameleons.

    SmartChameleons is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    SmartChameleons is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SmartChameleons; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""

#The MODEL
from Chameleon import Chameleon
import random
import math
from mappings import *
from RL_CON import AI
import pickle
from ActionGroup import ActionGroup

class Model:
    def __init__(self, parent, nChams, colorList, nCycles, worldXSize, worldYSize):
        # the environment
        self.nChams = nChams*3 #number of starting chameleons is supposed the same for each color
        self.chList = []
        self.colorList = colorList
        self.nCycles = nCycles
        self.worldXSize = worldXSize
        self.worldYSize = worldYSize
        self.parent = parent

        self.time = 0 #start the model time

    # objects
    def buildObjects(self):
        i = 0
        for color in self.colorList:
            for j in range(self.nChams/3):
                newCham = Chameleon(i+j,color,None,self.worldXSize,self.worldYSize)
                item = self.parent.canvas.create_image(
                    newCham.xPosition*36,newCham.yPosition*36,
                    image=self.parent.cham_images[color])
                newCham.graphItem = item #creating the graphic item (reference) for the chameleon.
                self.chList.append(newCham)
            i+=self.nChams/3

        brainFile = open("brain_30k.bra", 'rb') #loading the brain for the simulation
        brain = pickle.load(brainFile)
        self.AI = AI(brain)

            
    # actions        
    def buildActions(self):
        #defining the action 

        self.chameleonsActionGroup = ActionGroup ("chameleons")

        def moveChams(self):
            random.shuffle(self.chList)
            for aCham in self.chList:
                #if the chameleon is smart then perform a smart move. Otherwise a random move would be done
                if aCham.color == self.parent.smartColor:
                    state = self.checkState(aCham)
                    changes = self.AI.computeSmartMove(state)
                    aCham.move(changes)
                else :
                    aCham.move()

        def handleCollisions(self):
            #checks collisions handling the color changes
            for cham in self.chList:
                cham.justChanged = False
            i = 0
            for cham1 in self.chList:
                if cham1.hasntJustChanged():
                    for cham2 in self.chList[i+1:self.nChams*3]:
                        if cham2.hasntJustChanged():
                            if cham1.color != cham2.color and \
                                self.distance(cham1.xPosition,cham1.yPosition,
                                        cham2.xPosition,cham2.yPosition) ==0 :
                                self.changeColors(cham1,cham2)
                i += 1

        self.chameleonsActionGroup.actions = [moveChams, handleCollisions]


        self.checkActionGroup = ActionGroup ("check")
            
        def check(self):
            #checks if the Chameleons are all of the same color
            flag = True
            color = self.chList[0].color
            for chameleon in self.chList[1:]:
                if chameleon.color != color:
                    flag = False
                    break
            self.parent.onlyOneColor = flag

        self.checkActionGroup.action = check

        self.actionGroupList = ["ask_chameleons", "check_model"]



    def step(self):
        #performs a step of the model simulation. Executes each action of the model action list.
        #this entails asking the chameleons to move then checking if they're all of the same color 
        self.time += 1
        for action_type in self.actionGroupList :
            if action_type == "ask_chameleons" :
                for action in self.chameleonsActionGroup.actions :
                    action(self)
            elif action_type == "check_model" :
                self.checkActionGroup.action(self)




    def distance(self,x1,y1,x2,y2):
        #general 2 dimension distance calculation
        x1 = (x1+self.worldXSize) % self.worldXSize
        x2 = (x2+self.worldXSize) % self.worldXSize
        y1 = (y1+self.worldYSize) % self.worldYSize
        y2 = (y2+self.worldYSize) % self.worldYSize
        return math.sqrt((x1-x2)**2+(y1-y2)**2)

    def checkState(self,cham):
        #checks the state of the smart chameleon. The smart chameleon performs a move by checking its state looking around himself,
        #2 cells away in each of the 8 directions.
        #after evaluating the state, the BRAIN would suggest the move according to the POLICY created via REINFORCEMENT LEARNING.
        #the state is identified by a 5x5 matrix where each cell is equal to 1 if one or more ENEMY chameleons are present on the
        #relative WORLD cell; it has 0 otherwise (the cell is empty or occupied by chameleons of the same color)
        state=[0]*25
        for otherCham in self.chList:
            if cham.color != otherCham.color :
                distance = self.distance(cham.xPosition,cham.yPosition,
                                         otherCham.xPosition,otherCham.yPosition)
                if distance <3:
                    position = [cham.xPosition - otherCham.xPosition,cham.yPosition - otherCham.yPosition]
                    state[positionToBox(position)]=1                        
        return state

    def changeColors(self,cham1,cham2):
        #changes the colors of the chameleons after their collision. The color of the image in the canvas is also changed.
        colorList = self.colorList[:]
        colorList.remove(cham1.color)
        colorList.remove(cham2.color)
        otherColor = colorList[0]
        cham1.changeColor(otherColor)
        cham2.changeColor(otherColor)
        self.parent.canvas.itemconfigure(cham1.graphItem,image=self.parent.cham_images[otherColor])
        self.parent.canvas.itemconfigure(cham2.graphItem,image=self.parent.cham_images[otherColor])





